from __future__ import with_statement
from vcCommand import *
from vcHelpers.Application import *
from vcHelpers.Selection import *
from vcHelpers.Robot2 import *
import os.path, basicolp, vcMatrix, vcVector


DEFAULT_URI = ""
DEFAULT_LINE_SEPARATOR = '\n'
DEFAULT_COLUMN_SEPARATOR = ';'
DEFAULT_ROW_SEPARATOR = '\\n'

app = getApplication()
sim = getSimulation()

def first_state():
  global uri_prop,prop_RobotID,prop_WeldID,prop_MotionType,prop_Base,prop_Tool,prop_Thickness,\
  prop_Approach,prop_Button,tagFilter,tagType,pick_button,part,\
  B_tool,B_base,B_gunType,B_toolsignal,B_toolJointIndex,B_weldtime,B_approach,create_spots,\
  program,routine,controller,basename,toolname,basenames,toolnames,robotconfig,robotconfigs
  
  program, routine, controller, basename, toolname, basenames, toolnames, robotconfig,  robotconfigs = getSelectedRobotsData()

  cmd = getCommand()
  for p in cmd.Properties:
    deleteProperty(p)
  ### Read CSV ###
  i = 10
  uri_prop = createProperty(VC_URI,'Read CSV::CSV spots file')
  uri_prop.Value = DEFAULT_URI
  uri_prop.Group =i = i+10 

  prop_DefaultSequence = createProperty(VC_STRING,'Read CSV::Default Sequence')
  prop_DefaultSequence.Value = "SpotID,X,Y,Z,Rx,Ry,Rz"
  prop_DefaultSequence.Group =i = i+10

  prop_startingLineID = createProperty(VC_STRING,'Read CSV::Start Line Flag')
  prop_startingLineID.Value = 'Start'
  prop_startingLineID.Group =i = i+10

  prop_endingLineID = createProperty(VC_STRING,'Read CSV::End Line Flag')
  prop_endingLineID.Value = 'Empty Line'
  prop_endingLineID.Group =i = i+10  

  prop_Button = createProperty(VC_BUTTON,'Read CSV::Read Spots')
  prop_Button.OnChanged = read_spots_from_CSV_file
  prop_Button.Group =i = i+10 
    
  ### ### ###
  ###  Generate robot motion statement ###
  ### ### ###
  j = 10
  tagFilter = createProperty(VC_STRING,'Create Statement::Spot tag name filter' )
  tagFilter.Group = j = j+10

  tagType = createProperty(VC_STRING,'Create Statement::Spot tag type',VC_PROPERTY_STEP)
  tagType.StepValues = ['Frame', 'Geometry']
  tagType.Value = 'Frame'
  tagType.Group = j = j+10

  pick_button = createProperty(VC_BUTTON,'Create Statement::Select Workpiece In 3D-world')
  pick_button.OnChanged = pickComponent
  pick_button.Group = j = j+10

  part = createProperty('Ref<Component>','Create Statement::Workpiece')
  part.OnChanged = searchFrames
  part.Group = j = j+10 

  B_tool = createProperty(VC_STRING,'Create Statement::Tool',VC_PROPERTY_STEP)
  B_tool.Group = j = j+10
  B_tool.StepValues = toolnames
  B_tool.Value = toolname

  B_base = createProperty(VC_STRING, 'Create Statement::Base',VC_PROPERTY_STEP)
  B_base.Group = j = j+10
  B_base.StepValues = basenames
  B_base.Value = basename

  B_weldtime = createProperty(VC_REAL,'Create Statement::Weld Time')
  B_weldtime.Group = j = j+10
  B_weldtime.Value = 1

  B_toolsignal = createProperty(VC_INTEGER,'Create Statement::Tool Signal')
  B_toolsignal.Group = j = j+10
  B_toolsignal.Value = 51

  B_approach = createProperty(VC_REAL,'Create Statement::DefaultApproach')
  B_approach.Group = j = j+10
  B_approach.Value = -20.0
  
  ### AutoOrientatePoints Tab ###
  z_clearance = createProperty(VC_REAL,'Create Statement::AutoOrientatePoints-Z_Offset')
  z_clearance.Group = j = j+10
  z_clearance.Value = -10
  
  auto_rotate_enabled = createProperty(VC_BOOLEAN,'Create Statement::AutoOrientatePoints-Enabled')
  auto_rotate_enabled.Group = j = j+10
  auto_rotate_enabled.Value = True
  
  test_step = createProperty(VC_REAL,'Create Statement::AutoOrientatePoints-TestStepAngle')
  test_step.Group = j = j+10
  test_step.Value = 22.5
  
  auto_orientate_points = createProperty('Ref<Component>','Create Statement::AutoOrientatePoints-EOAT')
  auto_orientate_points.Group = j = j+10
  auto_orientate_points.IsVisible = True

  create_spots = createProperty(VC_BUTTON,'Create Statement::CreateSpots')
  create_spots.Group = j = j+10
  create_spots.OnChanged = createSpots
  
  clear_routine = createProperty(VC_BUTTON,'Create Statement::Clear Routine')
  clear_routine.Group = j = j+10
  clear_routine.OnChanged = clearRoutine

  ###
  ### CSV Export Tab
  ###
  k = 10
  C_uri_prop = createProperty(VC_URI,'Post Process::Export')
  C_uri_prop.Value = ''
  C_uri_prop.Group =k = k+10 
  C_select_robot_family = createProperty(VC_STRING,'Post Process::Robot',VC_PROPERTY_STEP)
  C_select_robot_family.Group = k = k+10  
  C_csv_export = createProperty(VC_BUTTON,'Post Process::Create Robot Program')
  C_csv_export.OnChanged = postProcessRobotProgram
  C_csv_export.Group = k = k + 10

  executeInActionPanel()

addState( first_state )

### ### Read CSV ### ###

def print_file(arg):
  try:
    uri = arg.Value[8:]
    file = open( uri , 'r' )
    print file.read()
    print 'read OK'
  except:
    print 'invalid file path'


def read_spots_from_CSV_file(arg):

  sel = sel = app.SelectionManager
  comp = sel.getSelection(VC_SELECTION_COMPONENT)[0]
  if not comp:
    print 'Select a target component first and then try again'
    return
  else:
    if comp.findBehavioursByType(VC_RSLEXECUTOR):
      print 'Select a workpiece component, current selection is Robot'
      return
  
  uri_prop = getProperty('Read CSV::CSV spots file')
  uri = uri_prop.Value[8:]
  #print 111,uri
  #uri = 'C:\\Users\\muhamJa1\\Downloads\\EasySpot_rework\\CSV_spot_simple.csv'
  file = open( uri , 'r' )
  content = file.read()
  separator = '\n'
  lines = content.split(DEFAULT_LINE_SEPARATOR)
  start_line_flag = 0
  for line in lines:
    #terminate reading of *.csv file at empty line.
    if not line.strip():
     return  
    
    #start reading *.csv file when [Start Line Flag] is read.
    new_line = line.split(',') 
    if new_line[0] ==  getProperty('Read CSV::Start Line Flag').Value:
     start_line_flag = 1
    if start_line_flag == 0:
     continue   
    
    # skip lines which are not according to default sequence.
    if len(new_line)<6 or new_line[1].isalpha()==True:
     continue  
    
    spotname = new_line[0]
    x = float(new_line[1])
    y = float(new_line[2])
    z = float(new_line[3])
    rx = float(new_line[4])
    ry = float(new_line[5])
    rz = float(new_line[6])
    print spotname,x,y,z,rx,ry,rz

    target = vcMatrix.new()
    vec = vcVector.new(x,y,z)
    target.P = vec       
    vec = vcVector.new(rx,ry,rz)      
    target.WPR = vec
    a = 100
    easyspot_frames = comp.getFeature('EasySpots_' + comp.Name)
    if not easyspot_frames:
      easyspot_frames = comp.RootFeature.createFeature(VC_TRANSFORM, 'EasySpots_' + comp.Name)
    fea_frame = easyspot_frames.createFeature(VC_FRAME, spotname.strip() )
    fea_frame.PositionMatrix = target
    
    weld_tag_prop = fea_frame.createProperty( VC_BOOLEAN, 'WELD_SPOT_FRAME' )
    weld_tag_prop.Value = True
    weld_tag_prop.Group = a
    a += 10
    fea_frame.rebuild()
    
  app.render()  

'''
########################### separator #############################
'''
def getSnappedComponent(cmd):
  '''
  set the picked target component in the part property
  '''
  global snap_cmd

  getProperty('Create Statement::Workpiece').Value = cmd.TargetNode.Component
  snap_cmd.OnTargetSet = None
  app.cancelCommand()
  print "Selected component: "+getProperty('Create Statement::Workpiece').Value.Name

def pickComponent(arg):
  '''
  call to pick a target component in the 3d-world
  '''
  global snap_cmd
  snap_cmd = app.findCommand('interactiveSnap')
  snap_cmd.execute()
  snap_cmd.OnTargetSet = getSnappedComponent

def getFrames(component):
  '''
  finds all spots in the target component
  '''
  frames = {}
  frames = traverseNodes(component, frames)
  return frames
  
def traverseNodes(node, feas):
  '''
  traverses the nodal tree
  '''
  feas = traverseFeas(node.RootFeature, feas, node)
  for inode in node.Children:
    feas = traverseNodes(inode, feas)
  return feas
  
def traverseFeas(fea, feas, node):
  '''
  traverses the feature tree to find spot features
  '''
  
  tagFilter = getProperty('Create Statement::Spot tag name filter')
  tagFilter = tagFilter.Value.lower()
  tagType = getProperty('Create Statement::Spot tag type') # Geo tag -mode  or  Frame -mode
  if tagType.Value == 'Frame': tagtype = VC_FRAME
  else:                        tagtype = VC_GEOMETRY
  for f in fea.Children:
    if f.Type == tagtype and tagFilter in f.Name.lower():
      feas[f.Name] = [f, node]
    feas = traverseFeas(f, feas, node)
  return feas  

def searchFrames(arg):
  '''
  calls getFrames to find all spots in the target component
  sorts the found spots and stores spots in a dictionary and the names in the CurrentSpots prop stepvalues
  '''
  ### sequence > searchFrames >> getFrames [traverseNodes > traverseFeas] >> natural_sort [alphanum_key > tryint]
  
  global frames, weld_frames
  part = getProperty('Create Statement::Workpiece')
  part = part.Value
  if not part:
    return
  frames = getFrames(part) # { [frame, node], ....}
  weld_frames = []
  frs = frames.keys()
  natural_sort(frs)
  weld_frames = frs
  print weld_frames

### NATURAL SORT ###
####################
'''
helper functions to do human like sorting
sorting vals like:
spot1, spot7, spot10
instead of alphabetical like:
spot1, spot10, spot7
'''
import re
def tryint(s):
    try:
        return int(s)
    except:
        return s
def alphanum_key(s):
    """ Turn a string into a list of string and number chunks.
        "z23a" -> ["z", 23, "a"]
    """
    return [ tryint(c) for c in re.split('([0-9]+)', s) ]
def natural_sort(l):
    """ Sort the given list in the way that humans expect.
    """
    l.sort(key=alphanum_key)
### NATURAL SORT ###
#################### 


def createSpots(arg):
  '''
  generates the actual robot statemets based on the user settings
  '''
  global app,program,routine,controller,part,weld_frames
  
  ref_node = app.findComponent(part.Value.Name)
  executor_ = program

  if not executor_.Program.Routines:
    spot_routine = executor_.Program.addRoutine("Easy_Spot")
    gun_routine = executor_.Program.addRoutine("Gun_Action")
    create_gun_routine(gun_routine)
  else:
    spot_routine = executor_.Program.Routines[0]   

  gun_routine = executor_.Program.findRoutine('Gun_Action')
  if not gun_routine:
    gun_routine = executor_.Program.addRoutine("Gun_Action")
    create_gun_routine(gun_routine)

  autoRotProp = getProperty('Create Statement::AutoOrientatePoints-EOAT')
  baseFrame = getProperty('Create Statement::Base').Value
  toolFrame = getProperty('Create Statement::Tool').Value
  base,tool = get_base_tool(baseFrame,toolFrame)
  eoat_clone = app.cloneComponent(autoRotProp.Value)

  for f in weld_frames: 
   ref_frame = ref_node.findFeature(f)
   spot_rotation = rotateThis(eoat_clone, ref_node, ref_frame, tool)### modification of created targets to avoid EOAT collission ###
   print spot_rotation
   mtx = vcMatrix.new()    
   mtx = ref_node.WorldPositionMatrix * ref_frame.NodePositionMatrix
   invBaseInWorld = basicolp.getBaseInWorld(base,controller)
   invBaseInWorld.invert()
   mtx = invBaseInWorld * mtx
   if spot_rotation == 0:  	mtx.rotateRelZ(spot_rotation)
   else: mtx.rotateRelZ(spot_rotation[0]) 
   addSpotWeldStatements(spot_routine,gun_routine,base,tool,mtx)
   mtx = 0

  app.deleteComponent(eoat_clone)

def get_base_tool(baseFrame,toolFrame):
  
  lul = baseFrame.strip().split()
  if lul[0] == '*NULL*':
    base = controller.Bases[0]
  else:  
   for b in controller.Bases:
    if b.Name == baseFrame:
     base = b 

  lul = toolFrame.strip().split()
  if lul[0] == '*NULL*':
    tool = controller.Tools[0]
  else:    
   for t in controller.Tools:
    if t.Name == toolFrame:
     tool = t

  return base,tool 

def rotateThis(eoat, ref_node, ref_frame, tool):
  '''
  uses the EOAT to rotate spots to an orientation where gun doesn't hit the part (if possible)
  '''
  global controller

  if not eoat or not getProperty('Create Statement::AutoOrientatePoints-Enabled').Value:
    #do not study the collision,, returning zero rotations for all
    collision_free_rot = 0 
  else:
    tcp_prop    = tool
    if tcp_prop == '*NULL*': 
      print 'Warning: Tool (tcp) not set. AutoOrientate cannot be calculated.'
      collision_free_rot = 0
      return collision_free_rot
    test_step       = getProperty( 'Create Statement::AutoOrientatePoints-TestStepAngle' )
    z_clearance     = getProperty('Create Statement::AutoOrientatePoints-Z_Offset')
    robot_loc       = controller.Component.WorldPositionMatrix
    eoatCachePos    = eoat.PositionMatrix
    toolifaces      = eoat.findBehavioursByType( VC_ONETOONEINTERFACE )
    eoatInvPos      = eoat.InverseWorldPositionMatrix
    tool_in_world   = tool.Node.WorldPositionMatrix * tool.PositionMatrix
    eoatcomp_to_tcp = eoatInvPos * tool_in_world
    eoatcomp_to_tcp.invert()
    step_angle = test_step.Value
    test_rots = [x*step_angle for x in range(int (360/step_angle) ) ]
    collision_free_rot = []
    detector = sim.newCollisionDetector()
    detector.NodeListA = [ (eoat, VC_NODELIST_INCLUDE, VC_NODELIST_TREE ) ]
    detector.NodeListB = [ (sim.World, VC_NODELIST_INCLUDE, VC_NODELIST_TREE ) , (eoat, VC_NODELIST_EXCLUDE, VC_NODELIST_TREE )]

    frame = ref_frame
    in_node = ref_node
    candidates = []
    for a in test_rots:
     mtx = in_node.WorldPositionMatrix * frame.NodePositionMatrix
     mtx.rotateRelZ( a )
     mtx.translateRel(0,0,z_clearance.Value)
     mtx = mtx * eoatcomp_to_tcp
     eoat.PositionMatrix = mtx
     eoat.update()
     app.render()
     hit = detector.testOneCollision(0.1)
     if not hit:
       candidates.append([a, eoat.WorldPositionMatrix])
    if candidates:
     closest = min( candidates, key = lambda x: ( x[1].P-robot_loc.P ).length() )
     collision_free_rot.append( closest[0] )
    else:
     collision_free_rot.append(0) #all collided,, use current
    eoat.PositionMatrix = eoatCachePos
    eoat.update()
    app.render()
    controller.update()
  
  return collision_free_rot 

def addSpotWeldStatements(spot_routine,gun_routine,base,tool,mtx):

  approach = getProperty('Create Statement::DefaultApproach').Value

  if approach:
    offset = vcMatrix.new()
    offset.P= mtx.P
    offset.WPR = mtx.WPR
    offset.translateRel(0,0,approach)
    approachStatement_ = spot_routine.addStatement(VC_STATEMENT_LINMOTION)
    approachStatement_.Properties[1].Value = base
    approachStatement_.Properties[2].Value = tool
    approachStatement_.Positions[0].PositionInReference = offset

  moveStatement_ = spot_routine.addStatement(VC_STATEMENT_LINMOTION)
  moveStatement_.Properties[1].Value = base
  moveStatement_.Properties[2].Value = tool
  moveStatement_.Positions[0].PositionInReference = mtx
  call_gun_routine = spot_routine.addStatement(VC_STATEMENT_CALL)
  call_gun_routine.Properties[1].Value = 'Gun_Action'

  if approach:
    retractStatement_ = spot_routine.addStatement(VC_STATEMENT_LINMOTION)
    retractStatement_.Properties[1].Value = base
    retractStatement_.Properties[2].Value = tool
    retractStatement_.Positions[0].PositionInReference = offset
    offset = 0

def create_gun_routine(gun_routine):

  val_gun_trigger_ON_OFF = getProperty('Create Statement::Tool Signal').Value
  val_weld_time = getProperty('Create Statement::Weld Time').Value   

  gun_trigger_ON  = gun_routine.addStatement(VC_STATEMENT_SETBIN)
  gun_trigger_ON.Properties[1].Value = val_gun_trigger_ON_OFF
  gun_trigger_ON.Properties[2].Value = True
  
  weld_time = gun_routine.addStatement(VC_STATEMENT_DELAY)
  weld_time.Properties[1].Value = val_weld_time
  
  gun_trigger_OFF = gun_routine.addStatement(VC_STATEMENT_SETBIN)
  gun_trigger_OFF.Properties[1].Value = val_gun_trigger_ON_OFF
  gun_trigger_OFF.Properties[2].Value = False

  weld_time = gun_routine.addStatement(VC_STATEMENT_DELAY)
  weld_time.Properties[1].Value = 0.2

def postProcessRobotProgram(arg):
  pass  